/** The Super class for simulation animation applets
 */

package eduni.simanim;

import java.applet.*;
import java.awt.*;
import eduni.simjava.*;

/**
 * The superclass for all simulation animations. New animations
 * should <tt>extend</tt> this and provide bodies for the methods
 * <tt>anim_layout()</tt> and <tt>anim_init()</tt>.
 * @version     1.0, 4 September
 * @author      Ross McNab
 */

public abstract class Anim_applet extends Applet implements Runnable {
  Thread simThread;
  // The AWT bits
  Panel controls;
  Sim_anim trace_out;
  Button runBut, stopBut, layoutBut, pauseBut;
  Label speedLabel;
  Scrollbar speedScroll;
  // The options
  int speed = 10;
  boolean paused = false;

  // Two functions that should be over-ridden in the sub-class
  // MUST be used toset up all the entities
  /** This method <b>must</b> be provided in the subclass, it should setup
   * all the simulation entities, and link their ports. The method should
   * never be called by the subclass itself, call <tt>anim_relayout()</tt>
   * instead.
   */
  public abstract void anim_layout();
  // CAN be used to add gui stuff
  /** This method can be overidden in the subclass, and used to setup
   * any GUI objects to be used for input to the animation.
   * The method is called once when the applet starts up.
   */
  public void anim_init() {}          

  /** This method should not be overidden in the subclass, use
   * <tt>anim_init()</tt> instead. */
  public final void init() {
    controls = new Panel();
    GridBagLayout gb = new GridBagLayout();
    controls.setLayout(gb);
    GridBagConstraints c = new GridBagConstraints();

    c.fill = GridBagConstraints.BOTH;
    c.weightx = 1.0;
    layoutBut = new Button("Layout");
    gb.setConstraints(layoutBut, c);
    controls.add(layoutBut);
    runBut = new Button("Run");
    gb.setConstraints(runBut, c);
    controls.add(runBut);
    pauseBut = new Button("Pause");
    gb.setConstraints(pauseBut, c);
    pauseBut.disable();
    controls.add(pauseBut);
    c.gridwidth = GridBagConstraints.REMAINDER; //end row
    stopBut = new Button("Stop");
    gb.setConstraints(stopBut, c);
    stopBut.disable();
    controls.add(stopBut);

    c.gridwidth = 1;
    speedLabel = new Label("Speed: "+speed);
    gb.setConstraints(speedLabel, c);
    controls.add(speedLabel);
    c.gridwidth = GridBagConstraints.REMAINDER; //end row
    speedScroll = new Scrollbar(Scrollbar.HORIZONTAL, speed,
                                1, speed, 10*speed);
    speedScroll.setLineIncrement(10);
    gb.setConstraints(speedScroll, c);
    controls.add(speedScroll);

    trace_out = new Sim_anim(this);
    this.setLayout(new BorderLayout(5,5));
    this.add("Center", trace_out);
    this.add("South", controls);
    anim_init();
  }


  /** This method should not be overidden in the subclass */
  public final void start() {
    anim_relayout();
  }

  /** This method should not be overidden in the subclass */
  public final void stop() {
    simThread = null;
  }

  /** Reinitialises all the animation entities, by calling
   * <tt>anim_layout()</tt>, then redraws them. This method
   * can be called by the subclass if a change to the entities
   * needs to be shown before the simulation starts. It should be
   * used instead of calling <tt>anim_layout()</tt> directly.
   */
  public void anim_relayout() {
    Sim_system.initialise(trace_out, simThread);
    Sim_system.set_auto_trace(false);
    anim_layout();
    trace_out.draw_all_static();
    trace_out.repaint();  
  }

  int get_speed() { return speed; }
  boolean get_paused() { return paused; }

  void startSim() {
    stopBut.enable();
    pauseBut.enable(); paused = false;
    runBut.disable();
    layoutBut.disable();
    if (simThread == null) {
      simThread = new Thread(this);
    }
    simThread.start();
  }


  void stopSim() {
    stopBut.disable();
    pauseBut.disable(); paused = false;
    runBut.enable();
    layoutBut.enable();
    simThread = null;
    trace_out.close();
    System.gc();        // Tidy up memory usage
  }

  /** This method can be overidden in the subclass to handle GUI events.
   * Warning, if overidden, ensure the last line is:<p>
   *<tt>return super.handleEvent(e);</tt><p>
   */
  public boolean handleEvent(Event e) {
    Object target = e.target;

    if (e.id == Event.ACTION_EVENT) {
      if (target == runBut) {
        startSim();
      } else if (target == layoutBut) {
        anim_relayout();
      } else if (target == stopBut) {
        simThread.stop();
        stopSim();
      } else if (target == pauseBut) {
        if(paused) {
          paused = false;
          pauseBut.setLabel("Pause");
        } else {
          paused = true;
          pauseBut.setLabel("Restart");
        }
      }
    } else if(target == speedScroll){
      speed = speedScroll.getValue();
      speedLabel.setText("Speed: "+speed);
    }
    return super.handleEvent(e);
  }

  /** This method should not be overidden in the subclass */
  public final void run() {
    System.out.println("Entering simThread run()");
    // layout the simulation
    anim_relayout();
    // Run
    Sim_system.run();
    System.out.println("Exiting simThread run()");
    System.gc();                   // Tidy up memory usage
    trace_out.animate(simThread);  // now run anim :-)
    stopSim();
  }
}







